import gtdynamics
import matplotlib.pyplot as plt
import numpy as np

from roboplot.base import (get_figure_and_axes, set_axes_equal, set_title,
                           show_option)
from roboplot.trajectory.three_d import plot_pose3


def _dfs(robot, link_name):
    """Perform Depth-First-Search to get the robot tree."""
    link = robot.link(link_name)

    # append key type so there is no overwriting
    link_name += "_linkkey"

    node = {}
    node[link_name] = {}

    joints = link.joints()
    for j in joints:
        childLink = robot.joint(j.name()).child()

        if childLink.name() == link.name():
            continue

        # append key type so there is no overwriting
        joint_name = j.name() + "_jointkey"

        node[link_name][joint_name] = _dfs(robot, childLink.name())

    return node


@show_option
def plot_robot(robot: gtdynamics.Robot,
               root_link: str,
               fignum: int = 0,
               title: str = ""):
    """
    Plot the robot joints with connected links. Allows for visualizing robot models.

    Args:
        robot: The robot model as defined in GTDynamics.
        root_link: The base link in the provided robot.
        fignum: The figure number to plot on.
        title: The title of the plot.
        show: Flag indicating whether to display the plot.
    """

    # Generate the tree of links and joints via a Depth-First Search.
    robotTree = _dfs(robot, root_link)

    # Plot the base link frame.
    base = robot.link(root_link)
    plot_pose3(base.bMcom(), fignum=fignum)

    def _preorder(tree, key):
        """
        Traverse robot tree in pre-order format to plot the joints.
        
        We traverse over links so we can get the connections between joints.

        Args:
            tree: The result of running DFS on the robot.
            key: The node key to start traversal from.
        """
        if "_linkkey" in key:
            link = robot.link(key.replace("_linkkey", ""))

            xs, ys, zs = [], [], []

            for joint in link.joints():
                # If the joint's parent is the root link, add a connection to it.
                if joint.parent().name() == root_link:
                    xs.append(base.bMcom().translation()[0])
                    ys.append(base.bMcom().translation()[1])
                    zs.append(base.bMcom().translation()[2])

                # Compute joint's wTj using the parent link's bMcom
                # and the relation between the link and its parent.
                wTj = joint.parent().bMcom() * joint.jMp().inverse()

                xs.append(wTj.translation()[0])
                ys.append(wTj.translation()[1])
                zs.append(wTj.translation()[2])
                plot_pose3(wTj, fignum=fignum, axis_length=0.05)

            plt.plot(xs, ys, zs, "-k")

        child_keys = tree[key].keys()
        for child_key in child_keys:
            _preorder(tree[key], child_key)

    # Run _preorder traversal to plot the links correctly
    _preorder(robotTree, root_link + "_linkkey")

    # Set the axes to be proportional
    set_axes_equal(fignum)

    # Set the figure and window titles
    fig = set_title(fignum, title)

    return fig


@show_option
def plot_links(robot: gtdynamics.Robot,
               root_link: str,
               connect_links=True,
               fignum: int = 0,
               title: str = ""):
    """
    Plot only the links of the `robot`.

    Args:
        robot: The robot model as defined in GTDynamics.
        root_link: The base link in the provided robot.
        fignum: The figure number to plot on.
        title: The title of the plot.
        show: Flag indicating whether to display the plot.
    """
    for link in robot.links():
        if link.name() == root_link:
            plot_pose3(link.bMcom(), fignum=fignum)
        else:
            plot_pose3(link.bMcom(), fignum=fignum, axis_length=0.05)

    if connect_links:
        for joint in robot.joints():
            p = joint.parent().bMcom().translation()
            c = joint.child().bMcom().translation()
            _, axes = get_figure_and_axes(fignum, projection='3d')
            line = np.append(p[np.newaxis], c[np.newaxis], axis=0)
            axes.plot(line[:, 0], line[:, 1], line[:, 2])

    # Set the axes to be proportional
    set_axes_equal(fignum)

    # Set the figure and window titles
    fig = set_title(fignum, title)



"""Classes and functions to plot sensor data."""

from typing import Iterable

import numpy as np
from matplotlib import pyplot as plt

from roboplot.base import get_figure_and_axes, multirow, set_title, show_option


@show_option
def plot_measurements(x,
                      y,
                      fignum=0,
                      xlabel="x",
                      ylabel="y",
                      color="b",
                      title="Measurements"):
    """
    Plot a set of measurements as a line graph.

    Args:
        x: The x-axis data.
        y: The y-axis data.
        fignum: The figure number to plot this on.
        xlabel: The label on the x-axis.
        ylabel: The label on the y-axis.
        color: Color of the line plot (default is blue).
        title: Title of the plot.
        show: Flag indicating whether to display the plot.

    Returns:
        A matplotlib Figure.
    """
    fig, axes = get_figure_and_axes(fignum)

    axes.plot(x, y, "tab:{}".format(color))
    axes.set(xlabel=xlabel, ylabel=ylabel)

    fig = set_title(fig, title)

    return fig


@show_option
def plot_multirow_measurements(x: Iterable,
                               y: Iterable,
                               fignum: int = 0,
                               title: str = "Multiple Measurements",
                               xlabels: Iterable[str] = None,
                               ylabels: Iterable[str] = None,
                               colors: Iterable = None,
                               row_titles: Iterable[str] = None,
                               normalize: bool = False):
    """
    Plot multiple measurements over multiple rows for easy viewing and comparisons.

    Args:
        x (Iterable): The x-axis data. Common to all y-axis data rows.
        y (Iterable): The y-axis data. This is a NxR matrix of R sets of N data points.
        fignum (int): The figure number to plot this on.
        title (str): Title of the plot.
        xlabels (Iterable[str]): The labels on the x-axis.
        ylabels (Iterable[str]): The labels on the y-axis.
        colors (Iterable): Colors of each of the line plots.
        row_titles (Iterable[str]): The title of each individual subplot.
        normalize (bool): Flag to scale both axes to be of similar range.
        show (bool): Flag indicating whether to display the plot.

    Returns:
        A matplotlib Figure.
    """
    fig = multirow(x,
                   y,
                   fignum=fignum,
                   xlabels=xlabels,
                   ylabels=ylabels,
                   title=title,
                   row_titles=row_titles,
                   colors=colors,
                   normalize=normalize,
                   show=plot_multirow_measurements.__wrapped__.show)

    return fig


def get_limits(lower, upper, padding=0.1):
    """
    Add padding to the lower and upper limits.

    Args:
        lower: Lower limit
        upper: Upper limit
        padding: The amount of padding to add as a fractional percentage [0, 1]
    """
    upper_padding = padding * upper

    if upper > 0:
        upper += upper_padding
    elif upper == 0:
        upper += padding
    else:
        upper -= upper_padding

    lower_padding = padding * lower
    if lower > 0:
        lower -= lower_padding
    elif lower == 0:
        lower -= padding
    else:
        lower += lower_padding

    return (lower, upper)


@show_option
def plot_imu_measurements(timestamps: np.ndarray,
                          angular_velocity: np.ndarray,
                          linear_acceleration: np.ndarray,
                          fignum: int = 0,
                          title: str = "IMU Measurement Data",
                          labels=('X Axis', 'Y Axis', 'Z Axis'),
                          colors=('r', 'g', 'b')):
    """
    Plot measurements from an IMU.

    Args:
        timestamps (numpy.ndarray): N-dimensional vector of time indices.
        angular_velocity (numpy.ndarray): Nx3 matrix of angular velocity measurements.
        linear_acceleration (numpy.ndarray): Nx3 matrix of linear acceleration measurements.
        fignum (int): The figure number to plot this on.
        title (str): Title of the plot.
        labels (Iterable[str]): The labels for each subplot.
        colors (Iterable): Colors of each of the line plots.
        show (bool): Flag indicating whether to display the plot.

    Returns:
        A matplotlib Figure.
    """
    fig, axes = plt.subplots(2, 3, num=fignum, constrained_layout=True)

    set_title(fig, title)

    y_min, y_max = 0, 0

    # plot angular velocity
    for i, (label, color) in enumerate(zip(labels, colors)):
        ax = axes[0][i]
        ax.plot(
            timestamps / 10**9,  # timestamps are in nanosecs
            angular_velocity[:, i],
            color=color)
        ax.set_title(label)
        ax.set_xlabel('Time (s)')
        ax.set_ylabel('Angular Velocity')

        y_min, y_max = get_limits(angular_velocity[:, i].min(),
                                  angular_velocity[:, i].max(),
                                  padding=2)
        ax.set_ylim(y_min, y_max)

    # plot linear acceleration
    for i, (label, color) in enumerate(zip(labels, colors)):
        ax = axes[1][i]
        ax.plot(
            timestamps / 10**9,  # timestamps are in nanosecs
            linear_acceleration[:, i],
            color=color)
        ax.set_title(label)
        ax.set_xlabel('Time (s)')
        ax.set_ylabel('Linear Acceleration (m/s²)')

        y_min, y_max = get_limits(linear_acceleration[:, i].min(),
                                  linear_acceleration[:, i].max(),
                                  padding=2)
        ax.set_ylim(y_min, y_max)

    return fig



"""Functions to plot sequences."""

from typing import Iterable

import numpy as np
from matplotlib import pyplot as plt

from roboplot.base import (get_figure_and_axes, set_axes_equal, set_labels,
                           set_title, show_option)


@show_option
def plot_contact_sequence(values, fignum=0, colors='rgby', fig_size=(5, 1)):
    """
    
    """
    fig, axes = plt.subplots(len(values.keys()),
                             num=fignum,
                             constrained_layout=True,
                             figsize=fig_size)
    for idx, (key, value) in enumerate(values.items()):
        value = np.asarray(value)
        sequence = np.where(value > 0)[0]

        # Plot underlying dots
        axes[idx].scatter(np.arange(value.shape[0]),
                          np.zeros(value.shape),
                          marker='8',
                          s=fig_size[0])

        # Plot colored squares where contact was made
        axes[idx].scatter(sequence,
                          np.zeros(sequence.shape),
                          marker='s',
                          color=colors[idx],
                          s=60 * fig_size[0])

        axes[idx].set_ylabel(key, rotation='horizontal', va='center')

        # Turn off all axis elements except for ylabel
        # axes[idx].set_axis_off()
        axes[idx].xaxis.set_visible(False)
        axes[idx].tick_params(left=False, labelleft=False)
        #remove background patch (only needed for non-white background)
        axes[idx].patch.set_visible(False)

        # remove the box around the plot
        axes[idx].spines["right"].set_visible(False)
        axes[idx].spines["top"].set_visible(False)
        axes[idx].spines["left"].set_visible(False)
        axes[idx].spines["bottom"].set_visible(False)

    return fig


@show_option
def plot_foot_contacts(contact_points: dict,
                       fignum: int = 0,
                       title: str = "Contact Points",
                       labels: Iterable = ("X axis", "Y axis", "Z axis"),
                       colors: Iterable = 'rgby',
                       color_legend: Iterable = "rgby"):
    """
    Plot trajectory of foot contact points.

    Args:
        contact_points (dict): A dict with keys as foot indices and Tx3 matrix as values,
            where T is the number of timesteps.
        fignum (int): The figure number to plot this on.
        title (str): Title of the plot.
        colors (Iterable): Colors of each of the line plots.
        show (bool): Flag indicating whether to display the plot.

    Returns:
        A matplotlib Figure.
    """
    fig, axes = get_figure_and_axes(fignum, projection='3d')

    for idx, points in enumerate(contact_points.values()):
        axes.scatter(points[:, 0],
                     points[:, 1],
                     points[:, 2],
                     color=colors[idx],
                     label=color_legend[idx])

    set_title(fig, title)
    set_axes_equal(fig)
    set_labels(fig, labels)

    axes.legend(loc="upper right")

    return fig



"""
Code to plot publication-quality bayes nets and factor graphs.
"""
from typing import Iterable, Tuple, Dict

import daft
import gtsam

from roboplot.base import show_option


def get_factors_and_edges(graph):
    factors = []
    edges = []
    for idx in range(graph.nrFactors()):
        name = "f{0}".format(idx)
        factor = dict(name=name, label="", x=2 * idx, y=1)
        factors.append(factor)

        f = graph.at(idx)
        keys = f.keys()
        for key in keys:
            node = "{0}{1}".format(gtsam.symbolChr(key),
                                   gtsam.symbolIndex(key))
            edge = (node, name)
            edges.append(edge)

    return factors, edges


def get_nodes(values):
    nodes = []
    keys = values.keys()
    for idx, key in enumerate(keys):
        name = "{0}{1}".format(gtsam.symbolChr(key), gtsam.symbolIndex(key))
        label = "{0}_{1}".format(chr(gtsam.symbolChr(key)),
                                 gtsam.symbolIndex(key))
        node = dict(name=name, label="${0}$".format(label), x=2 * idx, y=0)
        nodes.append(node)

    return nodes


def _plot_bayes_net(nodes: Iterable[dict], edges: Iterable[Tuple]):
    """
    Helper function to plot a Bayes Net using Daft.

    Args:
        nodes: List of dicts defining each node of the graph.
        edges: List of 2-tuples specifying the edges between nodes.

    Returns:
        A PGM object.
    """
    G = daft.PGM(directed=True)

    for node in nodes:
        G.add_node(node["name"],
                   content=node["label"],
                   x=node["x"],
                   y=node["y"],
                   aspect=1.5)

    # merge edges for common factors
    factors = {}
    for edge in edges:
        if edge[1] in factors.keys():
            factors[edge[1]].append(edge[0])
        else:
            factors[edge[1]] = [edge[0]]

    for _, edge in factors.items():
        # These are edges to unitary factors
        if len(edge) < 2:
            continue

        G.add_edge(edge[0], edge[1])

    return G


@show_option
def plot_bayes_net(graph: gtsam.NonlinearFactorGraph,
                   values: gtsam.Values,
                   title: str = "Bayes Net") -> daft.PGM:
    """
    Plot a Bayes Net.

    *Note* that this function does not perform checks for existence of edges or nodes.

    Args:
        graph (gtsam.NonlinearFactorGraph): A factor graph which will be converted to a bayes net.
        values (gtsam.Values): A values dict.
        title (str): Plot title.
        show (bool): Flag indicating whether to display the plot.

    Returns:
        A daft PGM object.
    """
    # Get the factor and the edges
    _, edges = get_factors_and_edges(graph)

    # Get all the nodes of the graph
    nodes = get_nodes(values)

    G = _plot_bayes_net(nodes, edges)

    G.figure.suptitle(title)
    G.figure.canvas.manager.set_window_title(title.lower())

    # Set the `show_func` function so that the decorator calls it correctly.
    plot_bayes_net.__wrapped__.show_func = G.show

    G.render()

    return G


def _plot_factor_graph(nodes: Iterable[Dict], factors: Iterable[Dict],
                       edges: Iterable[Tuple]) -> daft.PGM:
    """
    Helper function to plot a factor graph using Daft.

    Args:
        nodes: List of dicts defining each node of the graph.
        factors: List of dicts defining each factor.
        edges: List of 2-tuples specifying the edges between nodes.

    Returns:
        A PGM object.
    """
    G = daft.PGM(directed=False)

    for node in nodes:
        G.add_node(node["name"],
                   content=node["label"],
                   x=node["x"],
                   y=node["y"],
                   aspect=1)

    for factor in factors:
        G.add_node(factor["name"],
                   content=factor["label"],
                   x=factor["x"],
                   y=factor["y"],
                   shape="rectangle",
                   scale=0.4,
                   plot_params={"fc": "black"})

    for edge in edges:
        G.add_edge(edge[0], edge[1])

    return G


@show_option
def plot_factor_graph(graph: gtsam.NonlinearFactorGraph,
                      values: gtsam.Values,
                      title: str = "Factor Graph") -> daft.PGM:
    """
    Plot a Factor Graph.

    *Note* that this function does not perform checks for existence of edges or nodes.

    Args:
        graph (gtsam.NonlinearFactorGraph): A factor graph.
        values (gtsam.Values): A values dict.
        title (str): Plot title.
        show (bool): Flag indicating whether to display the plot.

    Returns:
        A daft PGM object.
    """
    # Get the factor and the edges
    factors, edges = get_factors_and_edges(graph)

    # Get all the nodes of the graph
    nodes = get_nodes(values)

    G = _plot_factor_graph(nodes, factors, edges)

    G.figure.suptitle(title)
    G.figure.canvas.manager.set_window_title(title.lower())

    # Set the `show_func` function so that the decorator calls it correctly.
    plot_factor_graph.__wrapped__.show_func = G.show

    G.render()

    return G



"""
Basic building blocks for plotting.
"""
import inspect
from operator import ge
import sys
from collections import namedtuple
from typing import Iterable, Union

import numpy as np
import wrapt
from matplotlib import pyplot as plt

# Ignore warnings from dependencies.
if not sys.warnoptions:
    import warnings
    warnings.simplefilter("ignore")


def argspec_factory(wrapped):
    """Update the method signature of `wrapped`."""
    argspec = inspect.getfullargspec(wrapped)

    # Hack to add the "show" arg to the function signature
    args_dict = argspec._asdict()
    args_dict["args"].append("show")
    if args_dict["defaults"]:
        args_dict["defaults"] = tuple([*args_dict["defaults"], False])

    argspec = namedtuple('FullArgSpec',
                         args_dict.keys())._make(args_dict.values())

    new_argspec = inspect.FullArgSpec(argspec.args, argspec.varargs,
                                      argspec.varkw, argspec.defaults,
                                      argspec.kwonlyargs,
                                      argspec.kwonlydefaults,
                                      argspec.annotations)
    return new_argspec


def _arguments(*args, show=False, **kwargs):
    """
    Function to extract the `show` parameter.

    If `show` is not passed, it defaults to False.
    """
    return show, args, kwargs


def show_option(func):
    """
    Decorator to provide a default `show` option in the parameter list.

    Usages example:
    ```
    @show_option
    def plot_func():
        ...

    plot_func(show=True)
    ```

    If a plotting function is attached to an object instantiated in the function,
    you can set it as the `show_func` via this neat trick:
    ```
    @show_option
    def plot_axis():
        _, axis = plt.figure(1)
        plot_axis2.__wrapped__.show_func = axis.show
    ```

    Alternatively, you can also nest functions
    ```
    @show_option
    def plot_stream():
        ...

    @show_option
    def plot_rivers():
        plot_stream(show=plot_rivers.__wrapped__.show)

    plot_rivers(show=True)
    ```
    """
    @wrapt.decorator(adapter=argspec_factory(func))
    def _show_option(func, instance, args, kwargs):

        show, args, kwargs = _arguments(*args, **kwargs)

        # Attach the boolean flag so it is available downstream.
        func.show = show
        # Call the function.
        ret = func(*args, **kwargs)
        if show:
            if hasattr(func, "show_func"):
                # Check if custom show function is attached.
                func.show_func()
            else:
                # default to matplotlib's plt.show()
                plt.show()

        return ret

    return _show_option(func)


def get_figure(func):
    """
    Decorator to get the figure if a figure number is passed in instead of a Figure object.
    The figure parameter is required to be the first non-keyword argument.
    """
    def wrapper(*args, **kwargs):
        fig = args[0]
        if isinstance(fig, int):
            fig = plt.figure(fig)
        elif not isinstance(fig, plt.Figure):
            raise ValueError("Argument fig is not a valid Figure")

        args = tuple([fig] + list(args[1:]))
        return func(*args, **kwargs)

    return wrapper


@get_figure
def set_axes_equal(fig: Union[int, plt.Figure]):
    """
    Make axes of 3D plot have equal scale so that spheres appear as spheres,
    cubes as cubes, etc..  This is one possible solution to Matplotlib's
    ax.set_aspect('equal') and ax.axis('equal') not working for 3D.

    Args:
      fignum (int): An integer representing the figure number for Matplotlib.
    """
    ax = fig.gca()

    #TODO(Varun) Support both 3d and 2d plots correctly
    limits = np.array([
        ax.get_xlim3d(),
        ax.get_ylim3d(),
        ax.get_zlim3d(),
    ])

    origin = np.mean(limits, axis=1)
    radius = 0.5 * np.max(np.abs(limits[:, 1] - limits[:, 0]))

    ax.set_xlim3d([origin[0] - radius, origin[0] + radius])
    ax.set_ylim3d([origin[1] - radius, origin[1] + radius])
    ax.set_zlim3d([origin[2] - radius, origin[2] + radius])

    return fig, ax


@get_figure
def set_labels(fig: Union[int, plt.Figure], labels: Iterable):
    """
    Set the labels for each axis of the figure.

    Args:
        fig: Matplotlib figure or figure number.
        labels: Sequence of 2/3 strings depending number of axes.
    """
    axes = fig.gca()
    if len(labels) == 3:
        axes.set_zlabel(labels[2])

    axes.set_xlabel(labels[0])
    axes.set_ylabel(labels[1])


@get_figure
def set_title(fig: Union[int, plt.Figure], title: str) -> plt.Figure:
    """
    Set the title to the figure and the window indicated by `fig`.

    Args:
        fig: Matplotlib figure or figure number.
        title: The title string to set.

    Return:
        The figure with the title and window title set.
    """
    fig.suptitle(title)
    fig.canvas.manager.set_window_title(title.lower())
    return fig


def get_figure_and_axes(fignum: int, projection=None):
    """
    Get a figure specified by fignum along with its axes.
    If the figure already has an axes, then return it else create the axes.

    Args:
        fignum: The figure number.
        projection: Type of projection. '3d' if we want 3D plots.

    Returns:
        A tuple of figure and axes.
    """
    fig = plt.figure(fignum)
    if not fig.axes:
        axes = fig.add_subplot(projection=projection)
    else:
        axes = fig.axes[0]

    return fig, axes


@show_option
def multirow(x,
             y,
             fignum=0,
             xlabels=None,
             ylabels=None,
             title="",
             row_titles=None,
             colors=("tab:blue", "tab:green", "tab:orange"),
             normalize=False) -> plt.Figure:
    """
    Make a multi-row plot.

    Useful for comparing values such as measurements from in different axes.

    Args:
        x (numpy.ndarray): Iterable of x-axis values, common to all y-axes.
        y (numpy.ndarray): Iterable of y-axis values where the first axis are the data points and the second axis is the number of rows.
        fignum (int): Matplotlib figure number.
        xlabels (iterable[string]): The labels to apply on each row's x-axis.
        ylabels (iterable[string]): The labels to apply on each row's y-axis.
        title (string): The title of the plot.
        row_titles (iterable[string]): The sub-title for each row.
        colors (iterable[string]): The color of the plot on each row.
        normalize (bool): Flag indicating whether to make the scale of the x and y axes equivalent.
        show (bool): Flag indicating whether to display the plot.
    """
    num_rows = y.shape[1]

    fig = plt.figure(fignum)
    axes = fig.subplots(num_rows, 1)

    assert num_rows == len(colors), \
        "Not enough colors provided, expected {1}, got {0}".format(len(colors),
                                                                   num_rows)

    for i in range(num_rows):
        axes[i].plot(x, y[:, i], color=colors[i])
        if row_titles:
            axes[i].set_title(row_titles[i])

        xlabel = xlabels[i] if xlabels else "x"
        ylabel = ylabels[i] if ylabels else "y"
        axes[i].set(xlabel=xlabel, ylabel=ylabel)

        if normalize:
            limits = np.array([axes[i].get_xlim(), axes[i].get_ylim()])

            origin = np.mean(limits, axis=1)
            radius = 0.5 * np.max(np.abs(limits[:, 1] - limits[:, 0]))

            # axes[i].set_xlim([origin[0] - radius, origin[0] + radius])
            axes[i].set_ylim([origin[1] - radius, origin[1] + radius])

    plt.subplots_adjust(left=None,
                        bottom=None,
                        right=None,
                        top=None,
                        wspace=None,
                        hspace=1.0)

    title = title if title else "figure"
    fig = set_title(fig, title)

    return fig



"""Plotting functions for 2D poses."""

import matplotlib.pyplot as plt
import numpy as np
from matplotlib import patches
from mpl_toolkits.mplot3d import Axes3D  # pylint: disable=unused-import

from roboplot.base import get_figure_and_axes, set_title, show_option


def plot_covariance_ellipse_2d(axes: plt.Axes,
                               origin: np.ndarray,
                               covariance: np.ndarray,
                               alpha: float = 0.5):
    """
    Plots a Gaussian as an uncertainty ellipse

    Based on Maybeck Vol 1, page 366
    k=2.296 corresponds to 1 std, 68.26% of all probability
    k=11.82 corresponds to 3 std, 99.74% of all probability

    Args:
        axes (matplotlib.axes.Axes): Matplotlib axes.
        origin (gtsam.Point2): The origin in the world frame.
        covariance (numpy.ndarray): The marginal covariance matrix of the 2D point
            which will be represented as an ellipse.
        alpha (float): Transparency value for the plotted surface in the range [0, 1].
    """
    k = 2.296
    e, s = np.linalg.eig(covariance[0:2, 0:2])
    s1 = s[0, 0]
    s2 = s[1, 1]

    def ellipse(a, b, c):
        q = np.arange(0, 2 * np.pi, 2 * np.pi / 25)
        ellipse_x = np.cos(q)
        ellipse_y = np.sin(q)

        points = a * ellipse_x + b * ellipse_y
        x = c[0] + points[0, :]
        y = c[1] + points[1, :]
        return x, y

    ex, ey = ellipse(
        np.sqrt(s1 * k) * e[:, 0],
        np.sqrt(s2 * k) * e[:, 2], origin[0:2])

    axes.line(ex, ey, alpha=alpha, cmap='hot')


def plot_point2_on_axes(axes: plt.Axes,
                        point: np.ndarray,
                        linespec: str,
                        covariance: np.ndarray = None):
    """
    Plot a 2D point on given axis `axes` with given `linespec`.

    Args:
        axes (matplotlib.axes.Axes): Matplotlib axes.
        point (gtsam.Point2): The point to be plotted.
        linespec (string): String representing formatting options for Matplotlib.
        covariance (numpy.ndarray): Marginal covariance matrix to plot the uncertainty of the estimation.
    """
    axes.plot([point[0]], [point[1]], linespec)
    if covariance is not None:
        plot_covariance_ellipse_2d(axes, point, covariance)


def plot_point2(point,
                fignum=0,
                linespec="g*",
                covariance=None,
                axis_labels=('X axis', 'Y axis')):
    """
    Plot a 2D point on given figure with given `linespec`.

    Args:
        point (gtsam.Point2): The point to be plotted.
        fignum (int): Integer representing the figure number to use for plotting.
        linespec (string): String representing formatting options for Matplotlib.
        covariance (numpy.ndarray): Marginal covariance matrix to plot the uncertainty of the estimation.
        axis_labels (iterable[string]): List of axis labels to set.

    Returns:
        fig: The matplotlib figure.

    """
    fig, axes = get_figure_and_axes(fignum)
    plot_point2_on_axes(axes, point, linespec, covariance)

    axes.set_xlabel(axis_labels[0])
    axes.set_ylabel(axis_labels[1])

    return fig


@show_option
def plot_2d_points(values,
                   fignum=0,
                   linespec="g*",
                   marginals=None,
                   title="2D Points",
                   axis_labels=('X axis', 'Y axis')):
    """
    Plots the Point2s in `values`, with optional covariances.
    Finds all the Point2 objects in the given Values object and plots them.
    If a Marginals object is given, this function will also plot marginal
    covariance ellipses for each point.

    Args:
        values (gtsam.Values): Values dictionary consisting of points to be plotted.
        fignum (int): Integer representing the figure number to use for plotting.
        linespec (string): String representing formatting options for Matplotlib.
        marginals (numpy.ndarray): Marginal covariance matrix to plot the
            uncertainty of the estimation.
        title (string): The title of the plot.
        axis_labels (iterable[string]): List of axis labels to set.
        show: Flag indicating whether to display the plot.
    """

    keys = values.keys()

    # Plot points and covariance matrices
    for key in keys:
        try:
            point = values.atPoint2(key)
            if marginals is not None:
                covariance = marginals.marginalCovariance(key)
            else:
                covariance = None

            fig = plot_point2(point,
                              fignum=fignum,
                              linespec=linespec,
                              covariance=covariance,
                              axis_labels=axis_labels)

        except RuntimeError:
            continue
            # I guess it's not a Point2

    set_title(fignum, title)


def plot_pose2_on_axes(axes, pose, axis_length=0.1, covariance=None, k=5.0):
    """
    Plot a 2D pose on given axis `axes` with given `axis_length`.

    Args:
        axes (matplotlib.axes.Axes): Matplotlib axes.
        pose (gtsam.Pose2): The pose to be plotted.
        axis_length (float): The length of the camera axes.
        covariance (numpy.ndarray): Marginal covariance matrix
                to plot the uncertainty of the estimation.
        k (float): Covariance scaling factor. Default 5.0
    """
    # get rotation and translation (center)
    gRp = pose.rotation().matrix()  # rotation from pose to global
    origin = pose.translation()

    # draw the camera axes
    x_axis = origin + gRp[:, 0] * axis_length
    line = np.append(origin[np.newaxis], x_axis[np.newaxis], axis=0)
    axes.plot(line[:, 0], line[:, 1], 'r-')

    y_axis = origin + gRp[:, 1] * axis_length
    line = np.append(origin[np.newaxis], y_axis[np.newaxis], axis=0)
    axes.plot(line[:, 0], line[:, 1], 'g-')

    if covariance is not None:
        pPp = covariance[0:2, 0:2]
        gPp = np.matmul(np.matmul(gRp, pPp), gRp.T)

        w, v = np.linalg.eig(gPp)

        angle = np.arctan2(v[1, 0], v[0, 0])
        e1 = patches.Ellipse(origin,
                             np.sqrt(w[0] * k),
                             np.sqrt(w[1] * k),
                             np.rad2deg(angle),
                             fill=False)
        axes.add_patch(e1)


def plot_pose2(pose,
               fignum=0,
               axis_length=0.1,
               covariance=None,
               axis_labels=('X axis', 'Y axis', 'Z axis')):
    """
    Plot a 2D pose on given figure with given `axis_length`.

    Args:
        pose (gtsam.Pose2): The pose to be plotted.
        fignum (int): Integer representing the figure number to use for plotting.
        axis_length (float): The length of the camera axes.
        covariance (numpy.ndarray): Marginal covariance matrix to plot
                the uncertainty of the estimation.
        axis_labels (iterable[string]): List of axis labels to set.
    """
    fig, axes = get_figure_and_axes(fignum)
    plot_pose2_on_axes(axes,
                       pose,
                       axis_length=axis_length,
                       covariance=covariance)

    axes.set_xlabel(axis_labels[0])
    axes.set_ylabel(axis_labels[1])

    return fig



from roboplot.trajectory.two_d import *
from roboplot.trajectory.three_d import *
from roboplot.trajectory.trajectory import *
from roboplot.graph import *
from roboplot.measurements import *
from roboplot.sequence import *
try:
    from roboplot.robot import *
except ImportError:
    print("GTDynamics unavailable. Robot plotting features not enabled.")



"""Plotting functions for 3D poses and points."""

from typing import Iterable

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
from mpl_toolkits.mplot3d import Axes3D
from numpy.core.function_base import linspace  # pylint: disable=unused-import

from roboplot.base import get_figure_and_axes, set_title, show_option


def ellipsoid(rx: float, ry: float, rz: float, n: int):
    """
    Numpy equivalent of Matlab's ellipsoid function.

    Args:
        rx (float): Radius of ellipsoid in X-axis.
        ry (float): Radius of ellipsoid in Y-axis.
        rz (float): Radius of ellipsoid in Z-axis.
        n (int): The granularity of the ellipsoid plotted.

    Returns:
        tuple[numpy.ndarray]: The points in the x, y and z axes to use for the surface plot.
    """
    u = np.linspace(0, 2 * np.pi, n + 1)
    v = np.linspace(0, np.pi, n + 1)
    x = -rx * np.outer(np.cos(u), np.sin(v)).T
    y = -ry * np.outer(np.sin(u), np.sin(v)).T
    z = -rz * np.outer(np.ones_like(u), np.cos(v)).T

    return x, y, z


def plot_covariance_ellipse_3d(axes: matplotlib.axes.Axes,
                               origin: np.ndarray,
                               covariance: np.ndarray,
                               scale: float = 1.0,
                               n: int = 8,
                               alpha: float = 0.5):
    """
    Plots a Gaussian as an uncertainty ellipse

    Based on Maybeck Vol 1, page 366
    k=2.296 corresponds to 1 std, 68.26% of all probability
    k=11.82 corresponds to 3 std, 99.74% of all probability

    Args:
        axes (matplotlib.axes.Axes): Matplotlib axes.
        origin (gtsam.Point3): The origin in the world frame.
        covariance (numpy.ndarray): The marginal covariance matrix of the 3D point
                which will be represented as an ellipse.
        scale (float): Scaling factor of the radii of the covariance ellipse.
        n (int): Defines the granularity of the ellipse. Higher values indicate finer ellipses.
        alpha (float): Transparency value for the plotted surface in the range [0, 1].
    """
    k = 11.82
    U, S, _ = np.linalg.svd(covariance)

    radii = k * np.sqrt(S)
    radii = radii * scale
    rx, ry, rz = radii

    # generate data for "unrotated" ellipsoid
    xc, yc, zc = ellipsoid(rx, ry, rz, n)

    # rotate data with orientation matrix U and center c
    data = np.kron(U[:, 0:1], xc) + np.kron(U[:, 1:2], yc) + \
        np.kron(U[:, 2:3], zc)
    n = data.shape[1]
    x = data[0:n, :] + origin[0]
    y = data[n:2 * n, :] + origin[1]
    z = data[2 * n:, :] + origin[2]

    axes.plot_surface(x, y, z, alpha=alpha, cmap='hot')


def plot_point3_on_axes(axes: matplotlib.axes.Axes,
                        point: np.ndarray,
                        linespec: str,
                        covariance: np.ndarray = None):
    """
    Plot a 3D point on given axis `axes` with given `linespec`.

    Args:
        axes (matplotlib.axes.Axes): Matplotlib axes.
        point (gtsam.Point3): The point to be plotted.
        linespec (string): String representing formatting options for Matplotlib.
        covariance (numpy.ndarray): Marginal covariance matrix to plot the uncertainty of the estimation.
    """
    axes.plot([point[0]], [point[1]], [point[2]], linespec)
    if covariance is not None:
        plot_covariance_ellipse_3d(axes, point, covariance)


def plot_point3(point: np.ndarray,
                fignum: int = 0,
                linespec: str = "g*",
                covariance: np.ndarray = None,
                axis_labels: Iterable[str] = ('X axis', 'Y axis', 'Z axis')):
    """
    Plot a 3D point on given figure with given `linespec`.

    Args:
        point (gtsam.Point3): The point to be plotted.
        linespec (string): String representing formatting options for Matplotlib.
        fignum (int): Integer representing the figure number to use for plotting.
        covariance (numpy.ndarray): Marginal covariance matrix to plot the uncertainty of the estimation.
        axis_labels (iterable[string]): List of axis labels to set.

    Returns:
        fig: The matplotlib figure.

    """
    fig, axes = get_figure_and_axes(fignum, projection='3d')

    plot_point3_on_axes(axes, point, linespec, covariance)

    axes.set_xlabel(axis_labels[0])
    axes.set_ylabel(axis_labels[1])
    axes.set_zlabel(axis_labels[2])

    return fig


@show_option
def plot_3d_points(values,
                   fignum=0,
                   linespec="g*",
                   marginals=None,
                   title="3D Points",
                   axis_labels: Iterable[str] = ('X axis', 'Y axis',
                                                 'Z axis')):
    """
    Plots the Point3s in `values`, with optional covariances.
    Finds all the Point3 objects in the given Values object and plots them.
    If a Marginals object is given, this function will also plot marginal
    covariance ellipses for each point.

    Args:
        values (gtsam.Values): Values dictionary consisting of points to be plotted.
        fignum (int): Integer representing the figure number to use for plotting.
        linespec (string): String representing formatting options for Matplotlib.
        marginals (numpy.ndarray): Marginal covariance matrix to plot
                the uncertainty of the estimation.
        title (string): The title of the plot.
        axis_labels: List of axis labels to set.
        show: Flag indicating whether to display the plot.
    """

    keys = values.keys()

    # Plot points and covariance matrices
    for key in keys:
        try:
            point = values.atPoint3(key)
            if marginals is not None:
                covariance = marginals.marginalCovariance(key)
            else:
                covariance = None

            fig = plot_point3(point,
                              fignum=fignum,
                              linespec=linespec,
                              covariance=covariance,
                              axis_labels=axis_labels)

        except RuntimeError:
            continue
            # I guess it's not a Point3

    set_title(fignum, title)


def plot_pose3_on_axes(axes, pose, axis_length=0.1, covariance=None):
    """
    Plot a 3D pose on given axis `axes` with given `axis_length`.

    Args:
        axes (matplotlib.axes.Axes): Matplotlib axes.
        point (gtsam.Point3): The point to be plotted.
        linespec (string): String representing formatting options for Matplotlib.
        covariance (numpy.ndarray): Marginal covariance matrix to plot the uncertainty of the estimation.
    """
    # get rotation and translation (center)
    gRp = pose.rotation().matrix()  # rotation from pose to global
    origin = pose.translation()

    # draw the camera axes
    x_axis = origin + gRp[:, 0] * axis_length
    line = np.append(origin[np.newaxis], x_axis[np.newaxis], axis=0)
    axes.plot(line[:, 0], line[:, 1], line[:, 2], 'r-')

    y_axis = origin + gRp[:, 1] * axis_length
    line = np.append(origin[np.newaxis], y_axis[np.newaxis], axis=0)
    axes.plot(line[:, 0], line[:, 1], line[:, 2], 'g-')

    z_axis = origin + gRp[:, 2] * axis_length
    line = np.append(origin[np.newaxis], z_axis[np.newaxis], axis=0)
    axes.plot(line[:, 0], line[:, 1], line[:, 2], 'b-')

    # plot the covariance
    if covariance is not None:
        # covariance matrix in pose coordinate frame
        pPp = covariance[3:6, 3:6]
        # convert the covariance matrix to global coordinate frame
        gPp = gRp @ pPp @ gRp.T
        plot_covariance_ellipse_3d(axes, origin, gPp)


@show_option
def plot_pose3(pose,
               fignum=0,
               axis_length=0.1,
               covariance=None,
               axis_labels=('X axis', 'Y axis', 'Z axis')):
    """
    Plot a 3D pose on given figure with given `axis_length`.

    Args:
        pose (gtsam.Pose3): 3D pose to be plotted.
        fignum (int): Integer representing the figure number to use for plotting.
        linespec (string): String representing formatting options for Matplotlib.
        covariance (numpy.ndarray): Marginal covariance matrix to plot the uncertainty of the estimation.
        axis_labels (iterable[string]): List of axis labels to set.
        show: Flag indicating whether to display the plot.

    Returns:
        fig: The matplotlib figure.
    """
    fig, axes = get_figure_and_axes(fignum, projection='3d')

    plot_pose3_on_axes(axes,
                       pose,
                       covariance=covariance,
                       axis_length=axis_length)

    axes.set_xlabel(axis_labels[0])
    axes.set_ylabel(axis_labels[1])
    axes.set_zlabel(axis_labels[2])

    return fig



try:
    from setuptools import setup, find_packages
except ImportError:
    from distutils.core import setup, find_packages

packages = find_packages()

setup(
    name='roboplot',
    description='Library of plotting functions for everything robotics',
    version='1.0.3',
    author="Varun Agrawal et. al.",
    author_email="varunagrawal@gatech.edu",
    license='Simplified BSD license',
    keywords="robotics plotting prototyping visualization",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/borglab/roboplot",
    python_requires=">=3.6",
    # https://pypi.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Education',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Operating System :: MacOS',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 3',
    ],
    packages=packages,
    platforms="any",
    install_requires=open("requirements.txt").readlines(),
)



"""Functions to plot trajectories."""

import gtsam
import matplotlib.pyplot as plt
import numpy as np
from matplotlib import patches
from mpl_toolkits.mplot3d import Axes3D  # pylint: disable=unused-import

from roboplot.base import (get_figure_and_axes, set_axes_equal, set_title,
                           show_option)
from roboplot.trajectory.three_d import plot_pose3_on_axes
from roboplot.trajectory.two_d import plot_pose2_on_axes


@show_option
def plot_trajectory(values,
                    fignum=0,
                    scale=1,
                    marginals=None,
                    title="Plot Trajectory",
                    axis_labels=('X axis', 'Y axis', 'Z axis'),
                    axes_equal=True,
                    azimuth=-64,
                    elevation=27):
    """
    Plot a complete 2D/3D trajectory using poses in `values`.

    Args:
        values (gtsam.Values): Values containing some Pose2 and/or Pose3 values.
        fignum (int): Integer representing the figure number to use for plotting.
        scale (float): Value to scale the poses by.
        marginals (gtsam.Marginals): Marginalized probability values of the estimation.
            Used to plot uncertainty bounds.
        title (string): The title of the plot.
        axis_labels (iterable[string]): List of axis labels to set.
        axis_equal (bool): Flag to equi-scale all the plotting axes. Default is True.
        azimuth (float): Azimuth of the plot. Default is -64.
        elevation (float): Elevation of the plot. Default is 27.
        show (bool): Flag indicating whether to display the plot.
    """
    fig, axes = get_figure_and_axes(fignum, projection='3d')

    axes.set_xlabel(axis_labels[0])
    axes.set_ylabel(axis_labels[1])
    axes.set_zlabel(axis_labels[2])

    axes.view_init(azim=azimuth, elev=elevation)

    # Plot 2D poses, if any
    poses = gtsam.utilities.allPose2s(values)
    for key in poses.keys():
        pose = poses.atPose2(key)
        if marginals:
            covariance = marginals.marginalCovariance(key)
        else:
            covariance = None

        plot_pose2_on_axes(axes,
                           pose,
                           covariance=covariance,
                           axis_length=scale)

    # Then 3D poses, if any
    poses = gtsam.utilities.allPose3s(values)
    for key in poses.keys():
        pose = poses.atPose3(key)

        # Get the covariance matrix
        if marginals:
            covariance = marginals.marginalCovariance(key)
        else:
            covariance = None

        plot_pose3_on_axes(axes,
                           pose,
                           covariance=covariance,
                           axis_length=scale)

    set_title(fig, title)

    if axes_equal:
        # Update the plot space to encompass all plotted points
        axes.autoscale()
        # Set the 3 axes equal
        set_axes_equal(fignum)

    return fig


@show_option
def plot_incremental_trajectory(values,
                                fignum=0,
                                start=0,
                                scale=1,
                                marginals=None,
                                time_interval=0.1,
                                title="Plot Trajectory",
                                axis_labels=('X axis', 'Y axis', 'Z axis'),
                                axes_equal=True):
    """
    Incrementally plot a complete 3D trajectory using poses in `values`.

    Useful for generating trajectory animations.

    Args:
        values (gtsam.Values): Values dict containing the poses.
        fignum (int): Integer representing the figure number to use for plotting.
        start (int): Starting index to start plotting from.
        scale (float): Value to scale the poses by.
        marginals (gtsam.Marginals): Marginalized probability values of the estimation.
            Used to plot uncertainty bounds.
        time_interval (float): Time in seconds to pause between each rendering.
            Used to create animation effect.
        show (bool): Flag indicating whether to display the plot.
    """
    fig, axes = get_figure_and_axes(fignum, projection='3d')

    axes.set_xlabel(axis_labels[0])
    axes.set_ylabel(axis_labels[1])
    axes.set_zlabel(axis_labels[2])

    # Plot 2D poses, if any
    poses = gtsam.utilities.allPose2s(values)
    keys = gtsam.KeyVector(poses.keys())

    for key in keys[start:]:
        pose = poses.atPose2(key)
        if marginals:
            covariance = marginals.marginalCovariance(key)
        else:
            covariance = None

        plot_pose2_on_axes(axes,
                           pose,
                           covariance=covariance,
                           axis_length=scale)

    # Plot 3D poses, if any
    poses = gtsam.utilities.allPose3s(values)
    keys = gtsam.KeyVector(poses.keys())

    for key in keys[start:]:
        if values.exists(key):
            pose_i = values.atPose3(key)

            # Get the covariance matrix
            if marginals:
                covariance = marginals.marginalCovariance(key)
            else:
                covariance = None

            plot_pose3_on_axes(axes,
                               pose_i,
                               covariance=covariance,
                               axis_length=scale)

    set_title(fig, title)

    if axes_equal:
        # Update the plot space to encompass all plotted points
        axes.autoscale()
        # Set the 3 axes equal
        set_axes_equal(fignum)

    # Pause for a fixed amount of seconds
    plt.pause(time_interval)

    return fig
